#include <stdlib.h>
#include <string.h>
#include "anamod.h"

#define COMPONENTCOOKIE 4723847
#define CHECKVALIDCOMPONENT(x) ANAMODCHECKVALID(x,COMPONENTCOOKIE,"component")

struct componentobject_ {
  int cookie;
  char *name;
  PetscErrorCode(*module)(AnaModNumericalProblem,AnalysisItem*,int*,PetscBool*);
  int dataid; AnalysisDataType type; PetscBool hasval;
} ;

#undef __FUNCT__
#define __FUNCT__ "CreateComponentObject"
PetscErrorCode CreateComponentObject(const char *name,componentobject *obj)
{
  componentobject nnew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct componentobject_),&nnew); CHKERRQ(ierr);
  nnew->cookie = COMPONENTCOOKIE;
  ierr = PetscStrallocpy(name,(char**)&(nnew->name)); CHKERRQ(ierr);
  *obj = nnew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DestroyComponentObject"
PetscErrorCode DestroyComponentObject(componentobject cmp)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDCOMPONENT(cmp);
  ierr = PetscFree(cmp->name); CHKERRQ(ierr);
  ierr = PetscFree(cmp); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComponentSetModule"
PetscErrorCode ComponentSetModule
(componentobject cmpt,AnalysisDataType type,int id,
 PetscErrorCode(*f)(AnaModNumericalProblem,AnalysisItem*,int*,PetscBool*))
{
  PetscFunctionBegin;
  CHECKVALIDCOMPONENT(cmpt);
  cmpt->type = type;
  cmpt->dataid = id;
  cmpt->module  = f;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComponentCompute"
PetscErrorCode ComponentCompute
(componentobject cmpt,AnaModNumericalProblem prob,
 AnalysisItem *res,int *reslen,PetscBool *success)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDCOMPONENT(cmpt);
  if (!cmpt->module) 
    *success = PETSC_FALSE;
  else {
    ierr = (cmpt->module)(prob,res,reslen,success); CHKERRQ(ierr);
  }
  cmpt->hasval = *success;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComponentRetrieve"
PetscErrorCode ComponentRetrieve
(componentobject cmpt,AnaModNumericalProblem prob,
 AnalysisItem *res,int *reslen,PetscBool *success)
{
  PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDCOMPONENT(cmpt);
  *success = PETSC_FALSE;
  switch (cmpt->type) {
  case ANALYSISINTEGER : {
      int ir;
      ierr = PetscObjectComposedDataGetInt
	((PetscObject)prob,cmpt->dataid,ir,flg); CHKERRQ(ierr);
      res->i =  ir; *success = flg;
      break;}
  case ANALYSISDOUBLE : {
      PetscReal rr;
      ierr = PetscObjectComposedDataGetReal
	((PetscObject)prob,cmpt->dataid,rr,flg); CHKERRQ(ierr);
      res->r =  rr; *success = flg;
      break;}
  case ANALYSISINTARRAY : {
      int *iir;
      ierr = PetscObjectComposedDataGetIntstar
	((PetscObject)prob,cmpt->dataid,iir,flg); CHKERRQ(ierr);
      res->ii =  iir; *success = flg;
      ierr = PetscObjectComposedDataGetInt
	((PetscObject)prob,cmpt->dataid,*reslen,flg); CHKERRQ(ierr);
      break;}
  case ANALYSISDBLARRAY : {
      PetscReal *rrr;
      ierr = PetscObjectComposedDataGetRealstar
	((PetscObject)prob,cmpt->dataid,rrr,flg); CHKERRQ(ierr);
      res->rr =  rrr; *success = flg;
      ierr = PetscObjectComposedDataGetInt
	((PetscObject)prob,cmpt->dataid,*reslen,flg); CHKERRQ(ierr);
      break;}
  default :
    SETERRQ(MPI_COMM_WORLD,1,"Can not retrieve array data yet");
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComponentGetType"
PetscErrorCode ComponentGetType(componentobject cmpt,AnalysisDataType *t)
{
  PetscFunctionBegin;
  CHECKVALIDCOMPONENT(cmpt);
  *t = cmpt->type;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComponentGetName"
PetscErrorCode ComponentGetName(componentobject cmpt,const char **name)
{
  PetscFunctionBegin;
  CHECKVALIDCOMPONENT(cmpt);
  *name = cmpt->name;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComponentGetId"
PetscErrorCode ComponentGetId(componentobject cmpt,int *id)
{
  PetscFunctionBegin;
  CHECKVALIDCOMPONENT(cmpt);
  *id = cmpt->dataid;
  PetscFunctionReturn(0);
}

