#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "petscmat.h"
#include "syspro.h"
#include "syspro_impl.h"

/*! \file compute.c \anchor computec
  \brief System/Anamod and NMD interface
 */

/*! \page anamodinterface Computational modules interface

  SysPro has a facility for computing or retrieving metadata about the
  numerical problems it deals with. The interface comprises
  - specification of preserved categories under preprocessor application
  with PreprocessorSetPreservedCategories().
  

  THE FOLLOWING FACILITY IS DISABLED

  Standard it comes with a dummy library
  \c libsysprocompute.a 
  of routines that simply say "failed to compute/retrieve data". If the user
  has an actual computation package (such as AnaMod), then that can be
  interfaced by providing implementations of the routines
  SysProComputeQuantity() and SyspProRetrieveQuantity().
*/

#undef __FUNCT__
#define __FUNCT__ "PreprocessorSetPreservedCategories"
PetscErrorCode PreprocessorSetPreservedCategories
(const char *preprocess,const char *cats)
{
  SystemPreprocessor pp; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SystemPreprocessorGetByName(preprocess,&pp); CHKERRQ(ierr);
  if (!pp->preserved) {
    ierr = PetscStrallocpy(cats,(char**)&pp->preserved); CHKERRQ(ierr);
  } else {
    size_t l1,l2; char *cnew;
    ierr = PetscStrlen(pp->preserved,&l1); CHKERRQ(ierr);
    ierr = PetscStrlen(cats,&l2); CHKERRQ(ierr);
    ierr = PetscMalloc((l1+l2+2)*sizeof(char),&cnew); CHKERRQ(ierr);
    ierr = sprintf(cnew,"%s,%s",pp->preserved,cats);
    ierr = PetscFree(pp->preserved); CHKERRQ(ierr);
    pp->preserved = cnew;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessorGetPreservedCategories"
PetscErrorCode PreprocessorGetPreservedCategories
(const char *preprocess,const char **cats)
{
  SystemPreprocessor pp; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SystemPreprocessorGetByName(preprocess,&pp); CHKERRQ(ierr);
  *cats = pp->preserved;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclarePreprocessorRequiredCategories"
/*!
  Indicate which metadata categories need to be computed for a 
  successful application of this preprocessor.

  Arguments:
  - \c name : name of the current preprocessor
  - \c required : comma-separated list of metadata categories
*/
PetscErrorCode DeclarePreprocessorRequiredCategories
(char *name,char *required)
{
  SystemPreprocessor pp;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = SystemPreprocessorGetByName(name,&pp); CHKERRQ(ierr);
  ierr = PetscStrallocpy(required,(char**)&(pp->required)); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#if 0
#undef __FUNCT__
#define __FUNCT__ "SysProComputeQuantity"
PetscErrorCode SysProComputeQuantity
(NumericalProblem problem,char *cat,char *mod,void *res,PetscBool *flg)
{
  PetscFunctionBegin;
  *flg = PETSC_FALSE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProRetrieveQuantity"
PetscErrorCode SysProRetrieveQuantity
(NumericalProblem problem,char *cat,char *mod,void *res,PetscBool *flg)
{
  PetscFunctionBegin;
  *flg = PETSC_FALSE;
  PetscFunctionReturn(0);
}
#endif
