#include <stdlib.h>
#include "petscpc.h"
#include "petscksp.h"
#include "anamod.h"
#include "linpc.h"

/*!
 * Set the parameter value for a pc;
 * this routine is only called for parametrised pcs
 */
#undef __FUNCT__
#define __FUNCT__ "set_blocked_sub_pc"
static PetscErrorCode set_blocked_sub_pc(int pcv)
{
  const char *opt,*val = new char[5]; int ierr;
  PetscFunctionBegin;

  if (pcv>0) {
    ierr = PetscOptionsSetValue("-sub_ksp_type","preonly"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_sub_ksp_type","preonly"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-sub_pc_type","ilu"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_sub_pc_type","ilu"); CHKERRQ(ierr);
    opt = "-sub_pc_factor_levels";
    sprintf((char*)val,"%d",pcv-1);
    ierr = PetscOptionsSetValue(opt,val); CHKERRQ(ierr);
    opt = "-ana_sub_pc_factor_levels";
    sprintf((char*)val,"%d",pcv-1);
    ierr = PetscOptionsSetValue(opt,val); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-sub_pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_sub_pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
  } else if (pcv==-1) {
    ierr = PetscOptionsSetValue("-sub_ksp_type","preonly"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_sub_ksp_type","preonly"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-sub_pc_type","lu"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_sub_pc_type","lu"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-sub_pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_sub_pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
  } else if (pcv<-1) {
    ierr = PetscOptionsSetValue("-sub_ksp_type","gmres"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_sub_ksp_type","gmres"); CHKERRQ(ierr);
    sprintf((char*)val,"%d",-pcv-1);
    ierr = PetscOptionsSetValue("-sub_ksp_max_it",val); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_sub_ksp_max_it",val); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-sub_pc_type","jacobi"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_sub_pc_type","jacobi"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-sub_pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_sub_pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
  } else SETERRQ1(PETSC_COMM_SELF,1,"No idea what pcv=%d means",pcv);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "pc_string"
PetscErrorCode pc_string(KSPType pct,int pcv,int pcvv,char **s)
{
  PetscBool flg; char *ss; int ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(250*sizeof(char),&ss); CHKERRQ(ierr);

  /* block jacobi */
  ierr = PetscStrcmp(pct,PCBJACOBI,&flg); CHKERRQ(ierr);
  if (!flg) {
    ierr = PetscStrcmp(pct,PCASM,&flg); CHKERRQ(ierr);
  }
  if (flg) {
    char *j;
    ierr = PetscMalloc(100*sizeof(char),&j); CHKERRQ(ierr);
    if (pcv>0)
      sprintf(j,"local method ILU(%d)",pcv-1);
    else if (pcv<-1)
      sprintf(j,"local CG, epslog=%d",-pcv-1);
    else if (pcv==-1)
      sprintf(j,"local method full LU");
    else SETERRQ(PETSC_COMM_SELF,1,"No idea what this pcv==0 means for blocked pc");
    sprintf(ss,"%s (%s; over %d procs)",pct,j,pcvv);
    ierr = PetscFree(j); CHKERRQ(ierr);
    goto exit;
  }

  /* ilu */
  ierr = PetscStrcmp(pct,PCILU,&flg); CHKERRQ(ierr);
  if (flg) {
    sprintf(ss,"%s (levels %d)",pct,pcv-1);
    goto exit;
  }

#if defined(PETSC_HAVE_SPAI)
  /* spai */
  ierr = PetscStrcmp(pct,PCSPAI,&flg); CHKERRQ(ierr);
  if (flg) {
    sprintf(ss,"%s (blocksize %d)",pct,pcv);
    goto exit;
  }
#endif

  sprintf(ss,"%s",pct);
 exit:
  *s = ss;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "pc_short_string"
PetscErrorCode pc_short_string(KSPType pct,int pcv,int pcvv,char **s)
{
  PetscBool flg; char *ss; int ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(250*sizeof(char),&ss); CHKERRQ(ierr);

  /* block jacobi */
  ierr = PetscStrcmp(pct,PCBJACOBI,&flg); CHKERRQ(ierr);
  if (!flg) {
    ierr = PetscStrcmp(pct,PCASM,&flg); CHKERRQ(ierr);
  }
  if (flg) {
    char *j;
    ierr = PetscMalloc(100*sizeof(char),&j); CHKERRQ(ierr);
    if (pcv>0)
      sprintf(j,"ilu%d",pcv-1);
    else if (pcv<-1)
      sprintf(j,"cg%d",-pcv-1);
    else if (pcv==-1)
      sprintf(j,"lu");
    else SETERRQ(PETSC_COMM_SELF,1,"No idea what this pcv==0 means for blocked pc");
    sprintf(ss,"%s-%s",pct,j);
    ierr = PetscFree(j); CHKERRQ(ierr);
    goto exit;
  }

  /* ilu */
  ierr = PetscStrcmp(pct,PCILU,&flg); CHKERRQ(ierr);
  if (flg) {
    sprintf(ss,"%s%d",pct,pcv-1);
    goto exit;
  }

#if defined(PETSC_HAVE_SPAI)
  /* spai */
  ierr = PetscStrcmp(pct,PCSPAI,&flg); CHKERRQ(ierr);
  if (flg) {
    sprintf(ss,"%s%d",pct,pcv);
    goto exit;
  }
#endif

  /* pcs without pcv : bs95, jacobi, none */
  ierr = PetscStrcmp(pct,PCBS95,&flg); CHKERRQ(ierr);
  if (flg) {
    sprintf(ss,"%s",pct);
    goto exit;
  }
  ierr = PetscStrcmp(pct,PCJACOBI,&flg); CHKERRQ(ierr);
  if (flg) {
    sprintf(ss,"%s",pct);
    goto exit;
  }
  ierr = PetscStrcmp(pct,PCNONE,&flg); CHKERRQ(ierr);
  if (flg) {
    sprintf(ss,"%s",pct);
    goto exit;
  }

  /* any remaining cases */
  sprintf(ss,"%s%d",pct,pcv);
 exit:
  *s = ss;
  PetscFunctionReturn(0);
}

#if 0
/*
 * PC extra setup and statistics gathering
 */
static PetscErrorCode set_total_blocks(PC pc,PCType pct,int n,IS structure)
{
  PetscBool flg; int ierr;
  PetscFunctionBegin;

  ierr = PetscStrcmp(pct,PCBJACOBI,&flg); CHKERRQ(ierr);
  if (flg) {
    int i,nsplits,*lengths; const PetscInt *splits;
    if ((int)structure) {
      ierr = ISGetSize(structure,&nsplits); CHKERRQ(ierr);
      ierr = ISGetIndices(structure,&splits); CHKERRQ(ierr);
      ierr = PetscMalloc((nsplits-1)*sizeof(int),&lengths); CHKERRQ(ierr);
      for (i=0; i<nsplits-1; i++) lengths[i] = splits[i+1]-splits[i];
      ierr = ISRestoreIndices(structure,&splits); CHKERRQ(ierr);
      ierr = PCBJacobiSetTotalBlocks(pc,nsplits-1,lengths); CHKERRQ(ierr);
      ierr = PetscFree(lengths); CHKERRQ(ierr);
    } else {
      ierr = PCBJacobiSetTotalBlocks(pc,n,PETSC_NULL); CHKERRQ(ierr);
    }
  }

  ierr = PetscStrcmp(pct,PCASM,&flg); CHKERRQ(ierr);
  if (flg) {
    if ((long int)structure!=0) {
      IS *isses; int i,nsplits; const PetscInt *splits;
      ierr = ISGetSize(structure,&nsplits); CHKERRQ(ierr);
      ierr = ISGetIndices(structure,&splits); CHKERRQ(ierr);
      ierr = PetscMalloc((nsplits-1)*sizeof(IS),&isses); CHKERRQ(ierr);
      for (i=0; i<nsplits-1; i++) {
	ierr = ISCreateStride
	  (MPI_COMM_SELF,
	   splits[i+1]-splits[i],splits[i],1,isses+i); CHKERRQ(ierr);
      }
      ierr = ISRestoreIndices(structure,&splits); CHKERRQ(ierr);
      ierr = PCASMSetLocalSubdomains(pc,nsplits-1,isses); CHKERRQ(ierr);
    } else {
      ierr = PCASMSetLocalSubdomains(pc,n,PETSC_NULL); CHKERRQ(ierr);
    }
  }
  PetscFunctionReturn(0);
}
#endif

#undef __FUNCT__
#define __FUNCT__ "ilu_stats_function"
static PetscErrorCode ilu_stats_function(PC pc,void *ctx)
{
  /*
  sans_info_block info = (sans_info_block)ctx;
  PetscErrorCode ierr;
  */
  PetscFunctionBegin;
  /*
  ierr = PCILUGetShift(pc,&fraction); CHKERRQ(ierr);
  ierr = matrix_statistics_set_manteuffel_fraction
  (stats,fraction); CHKERRQ(ierr);
  */
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "get_pc_stats_function"
PetscErrorCode get_pc_stats_function(PCType pct,int(**f)(PC,void*))
{
  PetscBool flg; int ierr;

  PetscFunctionBegin;
  *f = 0;
  ierr = PetscStrcmp(pct,PCILU,&flg); CHKERRQ(ierr);
  if (flg) {
    *f = &ilu_stats_function;}
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "set_preconditioner_base_matrix"
PetscErrorCode set_preconditioner_base_matrix
(PCType pct,Mat B,Mat *Buse)
{
  PetscBool flg; PetscErrorCode ierr;

  PetscFunctionBegin;
  *Buse = B;

  /*
   * For a bunch of preconditioners, we need to convert the B matrix
   */
  ierr = PetscStrcmp(pct,PCBS95,&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = MatConvert
      (B,MATMPIAIJ,MAT_INITIAL_MATRIX,Buse); CHKERRQ(ierr);
    goto done;
  }
  ierr = PetscStrcmp(pct,PCMUMPS,&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = MatConvert
      (B,"mumps",MAT_INITIAL_MATRIX,Buse); CHKERRQ(ierr);
    goto done;
  }
  ierr = PetscStrcmp(pct,PCSPOOLES,&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = MatConvert
      (B,"spooles",MAT_INITIAL_MATRIX,Buse); CHKERRQ(ierr);
    goto done;
  }
  ierr = PetscStrcmp(pct,PCSUPERLU,&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = MatConvert
      (B,"superlu_dist",MAT_INITIAL_MATRIX,Buse); CHKERRQ(ierr);
    goto done;
  }
  ierr = PetscStrcmp(pct,PCUMFPACK,&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = MatConvert
      (B,"umfpack",MAT_INITIAL_MATRIX,Buse); CHKERRQ(ierr);
    goto done;
  }
 done:
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SetPetscOptionsForPC"
PetscErrorCode SetPetscOptionsForPC(PC pc,PCType pct0, int pcv,int pcvv)
{
  PCType pct = pct0; PetscBool f,no_pcv; PetscErrorCode ierr;
  PetscFunctionBegin;

  /*
   * set pc type
   */
  no_pcv = PETSC_FALSE;

  ierr = PetscStrcmp(pct,PCLU,&f); CHKERRQ(ierr);
  if (f) {
    ierr = PetscOptionsSetValue("-pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_factor_shift_positive_definite",PETSC_NULL); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCSILU,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCILU; char levels[5];
    ierr = PetscOptionsSetValue("-mat_no_inode",PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_mat_no_inode",PETSC_NULL); CHKERRQ(ierr);
    sprintf(levels,"%d",pcv-1);
    ierr = PetscOptionsClearValue("-pc_factor_levels"); CHKERRQ(ierr);
    ierr = PetscOptionsClearValue("-ana_pc_factor_levels"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-pc_factor_levels",levels); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_factor_levels",levels); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-pc_factor_shift_positive_definite","1"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_factor_shift_positive_definite","1"); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCILU,&f); CHKERRQ(ierr);
  if (f) {
    char levels[5];
    ierr = PetscOptionsSetValue("-mat_no_inode",PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_mat_no_inode",PETSC_NULL); CHKERRQ(ierr);
    sprintf(levels,"%d",pcv-1);
    ierr = PetscOptionsClearValue("-pc_factor_levels"); CHKERRQ(ierr);
    ierr = PetscOptionsClearValue("-ana_pc_factor_levels"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-pc_factor_levels",levels); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_factor_levels",levels); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-pc_factor_shift_nonzero",".000000001"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_factor_shift_nonzero",".000000001"); CHKERRQ(ierr);
    /*
    ierr = PetscOptionsSetValue("-pc_factor_shift_positive_definite","0"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_factor_shift_positive_definite","0"); CHKERRQ(ierr);
    */
    goto set;
  }

#if defined(PETSC_HAVE_SPAI)
  ierr = PetscStrcmp(pct,PCSPAI,&f); CHKERRQ(ierr);
  if (f) {
    char val[5];
    sprintf(val,"%d",pcv);
    ierr = PetscOptionsSetValue("-pc_spai_block_size",val); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_spai_block_size",val); CHKERRQ(ierr);
    goto set;
  }
#endif

  ierr = PetscStrcmp(pct,PCBS95,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCILU; /* presumably the matrix has been converted elsewhere */
    no_pcv = PETSC_TRUE;
    ierr = PetscOptionsClearValue("-pc_factor_levels"); CHKERRQ(ierr);
    ierr = PetscOptionsClearValue("-ana_pc_factor_levels"); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCBJACOBI,&f); CHKERRQ(ierr);
  if (f) {
    ierr = PetscOptionsSetValue("-pc_bjacobi_blocks","6"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_bjacobi_blocks","6"); CHKERRQ(ierr);
    ierr = set_blocked_sub_pc(pcv); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCASM,&f); CHKERRQ(ierr);
  if (f) {
    ierr = PetscOptionsSetValue("-pc_asm_type","basic"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_asm_type","basic"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-pc_asm_blocks","6"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_asm_blocks","6"); CHKERRQ(ierr);
    ierr = set_blocked_sub_pc(pcv); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCRASM,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCASM; /* restricted asm is asm with the restrict option */
    ierr = PetscOptionsSetValue("-pc_asm_type","restrict"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_asm_type","restrict"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-pc_asm_blocks","6"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_asm_blocks","6"); CHKERRQ(ierr);
    ierr = set_blocked_sub_pc(pcv); CHKERRQ(ierr);
    goto set;
  }

  /*
   * Hypre variants
   */
  ierr = PetscStrcmp(pct,PCBOOMERAMG,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCHYPRE;
    no_pcv = PETSC_TRUE;
    ierr = PetscOptionsSetValue
      ("-pc_hypre_type","boomeramg"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_type","boomeramg"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-pc_hypre_boomeramg_tol","0.0"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_boomeramg_tol","0.0"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-pc_hypre_boomeramg_max_iter","1"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_boomeramg_max_iter","1"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-pc_hypre_boomeramg_relax_type_coarse","Gaussian-elimination"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_boomeramg_relax_type_coarse","Gaussian-elimination"); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCEUCLID,&f); CHKERRQ(ierr);
  if (f) {
    char n[3];
    pct = PCHYPRE;
    no_pcv = PETSC_TRUE;
    ierr = PetscOptionsSetValue
      ("-pc_hypre_type","euclid"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_type","euclid"); CHKERRQ(ierr);
    ierr = PetscMemzero(n,3*sizeof(char)); CHKERRQ(ierr);
    sprintf(n,"%d",pcv);
    ierr = PetscOptionsSetValue
      ("-pc_hypre_euclid_levels",n); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_euclid_levels",n); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCPARASAILS,&f); CHKERRQ(ierr);
  if (f) {
    char n[3];
    pct = PCHYPRE;
    no_pcv = PETSC_TRUE;
    ierr = PetscOptionsSetValue
      ("-pc_hypre_type","parasails"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_type","parasails"); CHKERRQ(ierr);
    ierr = PetscMemzero(n,3*sizeof(char)); CHKERRQ(ierr);
    sprintf(n,"%d",pcv);
    ierr = PetscOptionsSetValue
      ("-pc_hypre_parasails_nlevels",n); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue
      ("-ana_pc_hypre_parasails_nlevels",n); CHKERRQ(ierr);
    goto set;
  }

  ierr = PetscStrcmp(pct,PCPILUT,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCHYPRE;
    no_pcv = PETSC_TRUE;
    ierr = PetscOptionsSetValue("-pc_hypre_type","pilut"); CHKERRQ(ierr);
    ierr = PetscOptionsSetValue("-ana_pc_hypre_type","pilut"); CHKERRQ(ierr);
    goto set;
  }

  /*
   * Direct methods are all PCLU
   */
  ierr = PetscStrcmp(pct,PCSUPERLU,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCLU;
    no_pcv = PETSC_TRUE;
    goto set;
  }
  ierr = PetscStrcmp(pct,PCMUMPS,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCLU;
    no_pcv = PETSC_TRUE;
    goto set;
  }
  ierr = PetscStrcmp(pct,PCSPOOLES,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCLU;
    no_pcv = PETSC_TRUE;
    goto set;
  }
  ierr = PetscStrcmp(pct,PCUMFPACK,&f); CHKERRQ(ierr);
  if (f) {
    pct = PCLU;
    no_pcv = PETSC_TRUE;
    goto set;
  }

 set:
  ierr = PetscOptionsSetValue("-pc_type",pct); CHKERRQ(ierr);
  ierr = PetscOptionsSetValue("-ana_pc_type",pct); CHKERRQ(ierr);
  if (no_pcv) PetscFunctionReturn(0);

  PetscFunctionReturn(0);
}

